<?php
/**
 * The model file of approval module of ZenTaoPMS.
 *
 * @copyright   Copyright 2009-2020 青岛易软天创网络科技有限公司(QingDao Nature Easy Soft Network Technology Co,LTD, www.cnezsoft.com)
 * @license     ZPL(http://zpl.pub/page/zplv12.html) or AGPL(https://www.gnu.org/licenses/agpl-3.0.en.html)
 * @author      Qiyu Xie <xieqiyu@easycorp.ltd>
 * @package     approval
 * @version     $Id: model.php 5107 2020-09-09 09:46:12Z xieqiyu@easycorp.ltd $
 * @link        http://www.zentao.net
 */
class approvalModel extends model
{
    /**
     * Construct.
     *
     * @access public
     * @return void
     */
    public function __construct()
    {
        parent::__construct();
        $this->loadModel('approvalflow');
    }

    /**
     * Get approval by id.
     *
     * @param  int    $id
     * @access public
     * @return object
     */
    public function getByID($id)
    {
        return $this->dao->select('*')->from(TABLE_APPROVAL)->where('id')->eq($id)->fetch();
    }

    /**
     * Get pending review list.
     *
     * @param  string $objectType
     * @access public
     * @return array
     */
    public function getPendingReviews($objectType)
    {
        $pendingList = $this->dao->select('t2.objectID')->from(TABLE_APPROVALNODE)->alias('t1')
            ->leftJoin(TABLE_APPROVALOBJECT)->alias('t2')
            ->on('t2.approval = t1.approval')
            ->where('t2.objectType')->eq($objectType)
            ->andWhere('t1.account')->eq($this->app->user->account)
            ->andWhere('t1.status')->eq('doing')
            ->fetchPairs('objectID');

        unset($pendingList['']);

        return $pendingList;
    }

    /**
     * Get nodes to confirm.
     *
     * @param int $flowID
     * @param int $version
     * @access public
     * @return void
     */
    public function getNodesToConfirm($flowID, $version = 0, $module = '', $dataID = '')
    {
        $flow  = $this->approvalflow->getByID($flowID, $version);

        $nodes = !empty($flow->nodes) ? json_decode($flow->nodes) : array();

        $nodesToConfirm = $this->approvalflow->searchNodesToConfirm($nodes, $module, $dataID);
        return $nodesToConfirm;
    }

    /**
     * Get front node list.
     *
     * @param  string $objectType
     * @param  int    $object
     * @access public
     * @return void
     */
    public function getFrontNodeList($objectType, $object)
    {
        $approvalObject = $this->dao->select('*')->from(TABLE_APPROVALOBJECT)
            ->where('objectType')->eq($objectType)
            ->andWhere('objectID')->eq($object->id)
            ->orderBy('id_desc')
            ->fetch();

        $nodes = $this->dao->select('*')->from(TABLE_APPROVALNODE)
            ->where('approval')->eq($approvalObject->approval)
            ->orderBy('id_asc')
            ->fetchGroup('node', 'id');

        /* Group by code|type. */
        foreach($nodes as $code => $nodeGroup)
        {
            foreach($nodeGroup as $id => $node)
            {
                unset($nodes[$code][$node->id]);
                $nodes[$code][$node->type][] = $node;

                if($node->type != 'review') continue;

                $nodes[$code][$node->type]['users'][$node->account]['result'] = $node->result;
                $nodes[$code][$node->type]['users'][$node->account]['status'] = $node->status;
            }
        }

        /* Compute status and result. */
        foreach($nodes as $code => $nodeGroup)
        {
            foreach($nodeGroup as $type => $node)
            {
                if(isset($node['users']))
                {
                    foreach($node['users'] as $account => $resultAndStatus)
                    {
                        if($resultAndStatus['status'] == 'doing')
                        {
                            $nodes[$code][$type]['status'] = 'doing';
                            $nodes[$code][$type]['result'] = '';
                            break;
                        }
                        else
                        {
                            if($resultAndStatus['status'] == 'wait')
                            {
                                $nodes[$code][$type]['status'] = 'wait';
                                $nodes[$code][$type]['result'] = '';
                                break;
                            }
                            elseif($resultAndStatus['result'] == 'fail')
                            {
                                $nodes[$code][$type]['status'] = 'done';
                                $nodes[$code][$type]['result'] = 'fail';
                                break;
                            }
                            else
                            {
                                $nodes[$code][$type]['status'] = 'done';
                                $nodes[$code][$type]['result'] = 'pass';
                            }
                        }
                    }
                }
            }
        }

        return $nodes;
    }

    /**
     * Get approval object by id.
     *
     * @param  int    $approvalID
     * @access public
     * @return void
     */
    public function getApprovalObjectByID($approvalID)
    {
        return $this->dao->select('*')->from(TABLE_APPROVALOBJECT)->where('approval')->eq($approvalID)->fetch();
    }

    /**
     * Get approval id by object id.
     *
     * @param  int    $objectID
     * @param  string $objectType
     * @access public
     * @return void
     */
    public function getApprovalIDByObjectID($objectID, $objectType = 'review')
    {
        return $this->dao->select('approval')->from(TABLE_APPROVALOBJECT)->alias('t1')
            ->leftJoin(TABLE_APPROVAL)->alias('t2')->on('t1.approval=t2.id')
            ->where('t1.objectType')->eq($objectType)
            ->andWhere('t1.objectID')->eq($objectID)
            ->andWhere('t2.status')->eq('done')
            ->orderBy('t1.id asc')
            ->fetchPairs();
    }

    /**
     * Get last reviewDate.
     *
     * @param  int    $approvalID
     * @access public
     * @return string
     */
    public function getLastReviewDate($approvalID)
    {
        return $this->dao->select('reviewedDate')->from(TABLE_APPROVALNODE)
            ->where('approval')->eq($approvalID)
            ->andWhere('reviewedDate')->ne('0000-00-00 00:00:00')
            ->orderBy('reviewedDate_desc')
            ->limit(1)
            ->fetch('reviewedDate');
    }

    /**
     * Get node by id.
     *
     * @param  int    $nodeID
     * @access public
     * @return void
     */
    public function getNodeByID($nodeID)
    {
        return $this->dao->select('*')->from(TABLE_APPROVALNODE)->where('id')->eq($nodeID)->fetch();
    }

    /**
     * Get approval by object.
     *
     * @param  string $objectType
     * @param  int    $objectID
     * @access public
     * @return void
     */
    public function getByObject($objectType, $objectID)
    {
        return $this->dao->select('*')->from(TABLE_APPROVAL)
            ->where('objectType')->eq($objectType)
            ->andWhere('objectID')->eq($objectID)
            ->orderBy('id_desc')
            ->fetch();
    }

    /**
     * Get list by object.
     *
     * @param  string $objectType
     * @param  int    $objectID
     * @access public
     * @return array
     */
    public function getListByObject($objectType, $objectID)
    {
        return $this->dao->select('*')->from(TABLE_APPROVAL)
            ->where('objectType')->eq($objectType)
            ->andWhere('objectID')->eq($objectID)
            ->orderBy('id_desc')
            ->fetchAll('id');
    }

    /**
     * Get node reviewers.
     *
     * @param  int    $approvalID
     * @access public
     * @return void
     */
    public function getNodeReviewers($approvalID)
    {
        $nodeGroups = $this->dao->select('*')->from(TABLE_APPROVALNODE)->where('approval')->eq($approvalID)->orderBy('id')->fetchGroup('node', 'id');
        $nodeMap    = array();
        $this->loadModel('file');
        foreach($nodeGroups as $nodeID => $nodes)
        {
            $nodeMap[$nodeID] = array('reviewers' => array(), 'ccs' => array(), 'doing' => array(), 'status' => 'wait', 'result' => '');

            $nodeStatus = 'wait';
            $nodeResult = '';
            $isIgnore   = true;
            foreach($nodes as $node)
            {
                if($node->opinion) $node = $this->file->replaceImgURL($node, 'opinion');
                if($node->result == 'ignore' and $node->status == 'doing') $node->status = 'done';
                if($node->type == 'review') $nodeMap[$nodeID]['reviewers'][] = array('account' => $node->account, 'status' => $node->status, 'result' => $node->result, 'reviewedDate' => helper::isZeroDate($node->reviewedDate) ? '' : $node->reviewedDate, 'opinion' => $node->opinion);
                if($node->type == 'cc')     $nodeMap[$nodeID]['ccs'][]       = $node->account;

                if($node->status == 'doing')
                {
                    $isIgnore   = false;
                    $nodeStatus = 'doing';
                    $nodeResult = '';
                    $nodeMap[$node->node]['doing'][] = $node->account;
                }
                elseif($node->status == 'done' and $nodeStatus != 'doing')
                {
                    if($node->result != 'ignore') $isIgnore = false;

                    $nodeStatus = 'done';
                    if(!$isIgnore and $nodeResult != 'fail')     $nodeResult = 'success';
                    if(!$isIgnore and $node->result == 'fail')   $nodeResult = 'fail';
                    if($isIgnore  and $node->result == 'ignore') $nodeResult = 'ignore';
                }
            }

            $nodeMap[$nodeID]['status'] = $nodeStatus;
            $nodeMap[$nodeID]['result'] = $nodeResult;
        }

        /* Sort reviewers. Set status of done before doing. Set mine to head in the doing list. */
        foreach($nodeMap as $nodeID => $maps)
        {
            if(empty($maps['doing'])) continue;

            $reviewers = $maps['reviewers'];

            $ordered   = array();
            $mine      = array();
            $doings    = array();
            foreach($reviewers as $reviewer)
            {
                if($reviewer['status'] != 'doing')
                {
                    $ordered[] = $reviewer;
                    continue;
                }

                if(empty($mine) and $reviewer['account'] == $this->app->user->account)
                {
                    $mine = $reviewer;
                    continue;
                }

                $doings[] = $reviewer;
            }

            if($mine) $ordered[] = $mine;
            if($doings)
            {
                foreach($doings as $reviewer) $ordered[] = $reviewer;
            }

            $nodeMap[$nodeID]['reviewers'] = $ordered;
        }

        return $nodeMap;
    }

    /**
     * Get current reviewers.
     *
     * @param  int    $approvalID
     * @access public
     * @return array
     */
    public function getCurrentReviewers($approvalID)
    {
        return $this->dao->select('account')->from(TABLE_APPROVALNODE)->where('approval')->eq($approvalID)->andWhere('status')->eq('doing')->fetchPairs();
    }

    /**
     * Create approval.
     *
     * @param int    $flowID
     * @param array  $reviewers
     * @param int    $version
     * @param string $objectType
     * @param int    $objectID
     * @param int    $projectID
     * @access private
     * @return void
     */
    private function create($flowID, $reviewers, $version = 0, $objectType = '', $objectID = 0, $projectID = 0)
    {
        $flow  = $this->approvalflow->getByID($flowID, $version);
        $nodes = json_decode($flow->nodes);

        $upLevel = '';
        /* If I am a manager, use it firstly. */
        $parent = $this->dao->select('parent')->from(TABLE_DEPT)->where('manager')->eq($this->app->user->dept)->orderBy('grade')->fetch('parent');
        if(!$parent) $upLevel = $this->dao->select('manager')->from(TABLE_DEPT)->where('id')->eq($parent)->fetch('manager');

        /* If I am not a manager, use the manager of my dept. */
        if(!$upLevel) $upLevel = $this->dao->select('manager')->from(TABLE_DEPT)->where('id')->eq($this->app->user->dept)->fetch('manager');

        /* Get users of all roles. */
        $roles = $this->dao->select('id,users')->from(TABLE_APPROVALROLE)->where('deleted')->eq(0)->fetchPairs();
        foreach($roles as $id => $users)
        {
            $roles[$id] = explode(',', trim($users, ','));
        }

        /* Get users of project roles. */
        $projectRoles = array();
        if($projectID)
        {
            $projectRoleList = $this->dao->select('account,projectRole')->from(TABLE_TEAM)
                ->where('root')->eq($projectID)
                ->andWhere('type')->eq('project')
                ->andWhere('projectRole')->ne('')
                ->fetchAll();

            foreach($projectRoleList as $projectRole)
            {
                foreach(explode(',', $projectRole->projectRole) as $role)
                {
                    $projectRoles[$role][] = $projectRole->account;
                }
            }
        }

        /* Generate nodes. */
        $this->approvalflow->genNodes($nodes, array('reviewers' => $reviewers, 'upLevel' => $upLevel, 'roles' => $roles, 'projectRoles' => $projectRoles));

        if(dao::isError()) return false;

        /* Insert nodes. */
        $approval = new stdclass();
        $approval->flow        = $flowID;
        $approval->objectType  = $objectType;
        $approval->objectID    = $objectID;
        $approval->version     = $flow->version;
        $approval->createdBy   = $this->app->user->account;
        $approval->createdDate = helper::now();
        $approval->nodes       = json_encode($nodes);

        $this->dao->insert(TABLE_APPROVAL)->data($approval)->exec();
        $approvalID = $this->dao->lastInsertID();

        $this->insertNodes($approvalID, $nodes, array());

        return $approvalID;
    }

    /**
     * Create approval object for a object.
     *
     * @param  int    $root
     * @param  int    $objectID
     * @param  string $objectType
     * @param  array  $reviewers
     * @param  array  $ccers
     * @param  array  $nodeIdList
     * @param  string $type
     * @param  int    $projectID
     * @access public
     * @return string If finished, return pass|fail.
     */
    public function createApprovalObject($root = 0, $objectID = 0, $objectType = 0, $reviewers = array(), $ccers = array(), $nodeIdList = array(), $type = '', $projectID = 0)
    {
        /* Create approval. */
        if($objectType == 'review')
        {
            $approvalflowObject = $this->dao->select('*')->from(TABLE_APPROVALFLOWOBJECT)->where('root')->eq($root)->andWhere('objectType')->eq($type)->fetch();
            $flowID = $approvalflowObject ? $approvalflowObject->flow : $this->dao->select('*')->from(TABLE_APPROVALFLOW)->where('code')->eq('simple')->fetch()->id;
        }
        else
        {
            $approvalflowObject = $this->dao->select('*')->from(TABLE_APPROVALFLOWOBJECT)->where('root')->eq($root)->andWhere('objectType')->eq($objectType)->fetch();
            if(!$approvalflowObject) return false;
            $flowID = $approvalflowObject->flow;
        }

        $approvalUsers = array();
        foreach($nodeIdList as $id)
        {
            $approvalUsers[$id] = array('reviewers' => array(), 'ccs' => array());
            foreach($reviewers[$id] as $reviewer)
            {
                if($reviewer) $approvalUsers[$id]['reviewers'][] = $reviewer;
            }

            foreach($ccers[$id] as $ccer)
            {
                if($ccer) $approvalUsers[$id]['ccs'][] = $ccer;
            }
        }

        /* If submit a new approval, set old approval status as done. */
        $oldApproval = $this->getByObject($objectType, $objectID);
        if($oldApproval)
        {
            $this->dao->update(TABLE_APPROVAL)->set('status')->eq('done')->where('id')->eq($oldApproval->id)->exec();
            $this->dao->update(TABLE_APPROVALNODE)->set('status')->eq('done')->where('approval')->eq($oldApproval->id)->exec();
        }

        $approvalID = $this->create($flowID, $approvalUsers, 0, $objectType, $objectID, $projectID);

        $data = new stdclass();
        $data->approval   = $approvalID;
        $data->objectType = $objectType;
        $data->objectID   = $objectID;
        $this->dao->insert(TABLE_APPROVALOBJECT)->data($data)->exec();

        /* Run approval. */
        $this->next($approvalID, '');

        /* If the flow is finished, change status of approval. */
        $doing = $this->dao->select('*')->from(TABLE_APPROVALNODE)
            ->where('approval')->eq($approvalID)
            ->andWhere('status')->eq('doing')
            ->fetchAll();

        if(empty($doing))
        {
            $reject = $this->dao->select('*')->from(TABLE_APPROVALNODE)->where('approval')->eq($approvalID)->andWhere('result')->eq('fail')->fetchAll();

            $result = empty($reject) ? 'pass' : 'fail';
            $this->finish($approvalID, $result);

            return array('result' => $result, 'approvalID' => $approvalID);
        }

        return array('result' => '', 'approvalID' => $approvalID);
    }

    /**
     * Insert node.
     *
     * @param int    $approvalID
     * @param object $node
     * @param array  $prevs
     * @access private
     * @return void
     */
    private function insertNode($approvalID, $node, $prevs)
    {
        $newPrev = '';
        if(isset($node->reviewers) and !empty($node->reviewers))
        {
            foreach($node->reviewers as $reviewer)
            {
                foreach($reviewer->users as $user)
                {
                    $data = new stdclass();
                    $data->approval     = $approvalID;
                    $data->node         = $node->id;
                    $data->title        = $node->title;
                    $data->type         = 'review';
                    $data->prev         = implode(',', $prevs);
                    $data->account      = $user;
                    $data->multipleType = $node->multiple;
                    $data->reviewType   = isset($node->reviewType) ? $node->reviewType : 'manual';

                    $this->dao->insert(TABLE_APPROVALNODE)->data($data)->exec();
                    $newPrev = $node->id;
                }
            }
        }

        if(isset($node->ccs) and !empty($node->ccs))
        {
            foreach($node->ccs as $cc)
            {
                foreach($cc->users as $user)
                {
                    $data = new stdclass();
                    $data->approval     = $approvalID;
                    $data->node         = $node->id;
                    $data->title        = $node->title;
                    $data->type         = 'cc';
                    $data->prev         = implode(',', $prevs);
                    $data->account      = $user;
                    $data->multipleType = $node->multiple;
                    $data->reviewType   = isset($node->reviewType) ? $node->reviewType : 'manual';

                    $this->dao->insert(TABLE_APPROVALNODE)->data($data)->exec();
                    $newPrev = $node->id;
                }
            }
        }

        return $newPrev;
    }

    /**
     * Insert nodes.
     *
     * @param  int   $approvalID
     * @param  array $nodes
     * @param  array $prevs
     * @access public
     * @return array
     */
    public function insertNodes($approvalID, $nodes, $prevs)
    {
        foreach($nodes as $node)
        {
            if($node->type == 'branch')
            {
                $nextPrevs = array();
                foreach($node->branches as $branch)
                {
                    $newPrevs = $this->insertNodes($approvalID, $branch->nodes, $prevs);
                    foreach($newPrevs as $newPrev)
                    {
                        if($newPrev) $nextPrevs[] = $newPrev;
                    }
                }
                if(!empty($nextPrevs)) $prevs = $nextPrevs;
            }
            else
            {
                $newPrev = $this->insertNode($approvalID, $node, $prevs);
                if($newPrev) $prevs = array($newPrev);
            }
        }

        return $prevs;
    }

    /**
     * Next viewer node.
     *
     * @param int    $approvalID
     * @param string $prev
     * @access private
     * @return void
     */
    private function next($approvalID, $prev)
    {
        /* Get next nodes. */
        if(!$prev)
        {
            $nodes = $this->dao->select('*')->from(TABLE_APPROVALNODE)->where('approval')->eq($approvalID)->andWhere('prev')->eq('')->fetchAll();
        }
        else
        {
            $nodes = $this->dao->select('*')->from(TABLE_APPROVALNODE)->where('approval')->eq($approvalID)->andWhere('prev')->like("%$prev%")->fetchAll();
        }
        if(empty($nodes)) return;

        $nodeMap = array();
        foreach($nodes as $index => $node)
        {
            if($prev)
            {
                $undone = $this->dao->select('*')->from(TABLE_APPROVALNODE)->where('approval')->eq($approvalID)->andWhere('node')->in($node->prev)->andWhere('status')->ne('done')->fetchAll();
                if(!empty($undone)) continue;   // Only all prevs are done, the node can flow.
            }

            if(!isset($nodeMap[$node->node])) $nodeMap[$node->node] = array('review' => array(), 'cc' => array());
            $nodeMap[$node->node][$node->type][] = $node->account;
            $nodeMap[$node->node]['reviewType']  = $node->reviewType;
        }

        foreach($nodeMap as $nodeID => $users)
        {
            if(empty($users['review']))
            {
                $this->cc($approvalID, $nodeID, $users['cc']);
                $this->next($approvalID, $nodeID);
            }
            else if($users['reviewType'] == 'pass') // Auto pass
            {
                $this->dao->update(TABLE_APPROVALNODE)
                    ->set('status')->eq('done')
                    ->set('result')->eq('pass')
                    ->where('approval')->eq($approvalID)
                    ->andWhere('node')->eq($nodeID)
                    ->exec();
                $this->cc($approvalID, $nodeID, $users['cc']);
                $this->next($approvalID, $nodeID);
            }
            else if($users['reviewType'] == 'reject') // Auto reject
            {
                $this->dao->update(TABLE_APPROVALNODE)
                    ->set('status')->eq('done')
                    ->set('result')->eq('fail')
                    ->where('approval')->eq($approvalID)
                    ->andWhere('node')->eq($nodeID)
                    ->andWhere('type')->eq('review')
                    ->exec();
                $this->dao->update(TABLE_APPROVALNODE)
                    ->set('status')->eq('done')
                    ->set('result')->eq('ignore')
                    ->where('approval')->eq($approvalID)
                    ->andWhere('status')->eq('wait')
                    ->exec();
            }
            else
            {
                $this->review($approvalID, $nodeID, $users['review']);
            }
        }
    }

    /**
     * review
     *
     * @param int    $approvalID
     * @param string $nodeID
     * @param array  $reviewers
     * @access private
     * @return void
     */
    private function review($approvalID, $nodeID, $reviewers)
    {
        $this->dao->update(TABLE_APPROVALNODE)->set('status')->eq('doing')->where('approval')->eq($approvalID)->andWhere('node')->eq($nodeID)->andWhere('type')->eq('review')->exec();
        $this->sendMessage($reviewers, $approvalID, $nodeID, 'review');
    }

    /**
     * cc
     *
     * @param int    $approvalID
     * @param string $nodeID
     * @param array  $ccs
     * @access private
     * @return void
     */
    private function cc($approvalID, $nodeID, $ccs)
    {
        $this->dao->update(TABLE_APPROVALNODE)->set('status')->eq('done')->where('approval')->eq($approvalID)->andWhere('node')->eq($nodeID)->exec();
        $this->sendMessage($ccs, $approvalID, $nodeID, 'cc');
    }

    /**
     * Send message.
     *
     * @param  array  $sendList
     * @param  int    $approvalID
     * @param  int    $nodeID
     * @param  string $method
     * @access private
     * @return void
     */
    private function sendMessage($sendList, $approvalID, $nodeID, $method)
    {
        $approval = $this->dao->select('t1.objectType, t1.objectID, t2.type')->from(TABLE_APPROVAL)->alias('t1')->leftJoin(TABLE_APPROVALFLOW)->alias('t2')->on('t1.flow = t2.id')->where('t1.id')->eq($approvalID)->fetch();
        $node     = $this->dao->select('*')->from(TABLE_APPROVALNODE)->where('approval')->eq($approvalID)->andWhere('node')->eq($nodeID)->fetch();

        $objectType = $approval->objectType;
        $objectID   = $approval->objectID;

        if(empty($objectID) or empty($objectType)) return;

        $this->loadModel('mail');
        $this->loadModel('message');

        /* Message config. */
        $messageSetting = $this->config->message->setting;
        if(is_string($messageSetting)) $messageSetting = json_decode($messageSetting, true);

        if(isset($messageSetting['mail']))
        {
            $actions = $messageSetting['mail']['setting'];
            if(!isset($actions['approval'])) return false;

            $actionType = $method == 'review' ? 'approval' : $method;
            if(!in_array($actionType, $actions['approval'])) return false;
        }

        if($approval->type == 'workflow')
        {
            $flow   = $this->dao->select('name, `table`')->from(TABLE_WORKFLOW)->where('module')->eq($objectType)->fetch();
            $object = $this->dao->select('*')->from($flow->table)->where('id')->eq($objectID)->fetch();
        }
        else
        {
            $suffix = '';
            $object = $this->loadModel($objectType)->getByID($objectID);
            if($objectType == 'review') $suffix = empty($object->project) ? '' : ' - ' . $this->loadModel('project')->getById($object->project)->name;
        }

        /* Load module and get vars. */
        $this->loadModel('action');
        $users      = $this->loadModel('user')->getPairs('noletter');
        $actions    = $this->action->getList($objectType, $objectID);
        $nameFields = $this->config->action->objectNameFields[$objectType];
        $title      = zget($object, $nameFields, '');
        $subject    = strtoupper($objectType) . ' #' . $object->id . ($approval->type == 'workflow' ? ' - ' . $flow->name : ' ' . $title . $suffix);
        $domain     = zget($this->config->mail, 'domain', common::getSysURL());

        foreach($actions as $action)
        {
            $action->appendLink = '';
            if(strpos($action->extra, ':') !== false)
            {
                list($extra, $id) = explode(':', $action->extra);
                $action->extra    = $extra;
                if($title) $action->appendLink = html::a($domain . helper::createLink($action->objectType, 'view', "id=$id", 'html'), "#$id " . $title);
            }
        }

        /* Get mail content. */
        $modulePath = $this->app->getModulePath($appName = '', 'approval');
        $oldcwd     = getcwd();
        $viewFile   = $modulePath . 'view/sendmail.html.php';
        chdir($modulePath . 'view');
        if(file_exists($modulePath . 'ext/view/sendmail.html.php'))
        {
            $viewFile = $modulePath . 'ext/view/sendmail.html.php';
            chdir($modulePath . 'ext/view');
        }
        ob_start();
        include $viewFile;
        foreach(glob($modulePath . 'ext/view/sendmail.*.html.hook.php') as $hookFile) include $hookFile;
        $mailContent = ob_get_contents();
        ob_end_clean();
        chdir($oldcwd);

        $toList = '';
        $ccList = '';
        if(is_array($sendList))
        {
            if($method == 'cc')     $ccList = implode(',', $sendList);
            if($method == 'review') $toList = implode(',', $sendList);
        }

        $this->mail->send($toList, $subject, $mailContent, $ccList);
        if($this->mail->isError()) error_log(join("\n", $this->mail->getError()));
    }

    /**
     * Pass approval.
     *
     * @param  string $objectType
     * @param  object $object
     * @param  string $extra
     * @access public
     * @return bool
     */
    public function pass($objectType, $object, $extra = '')
    {
        $approvalID = $this->dao->select('approval')->from(TABLE_APPROVALOBJECT)
            ->where('objectType')->eq($objectType)
            ->andWhere('objectID')->eq($object->id)
            ->orderBy('id_desc')
            ->fetch('approval');

        $data = new stdclass();
        $data->status       = 'done';
        $data->result       = 'pass';
        $data->date         = $object->createdDate;
        $data->opinion      = $object->opinion;
        $data->extra        = $extra;
        $data->reviewedBy   = $this->app->user->account;
        $data->reviewedDate = helper::now();

        $nodes = $this->dao->select('*')->from(TABLE_APPROVALNODE)
            ->where('approval')->eq($approvalID)
            ->andWhere('status')->eq('doing')
            ->andWhere('account')->eq($this->app->user->account)
            ->fetchAll();
        foreach($nodes as $node)
        {
            $this->dao->update(TABLE_APPROVALNODE)->data($data)->where('id')->eq($node->id)->exec();
            if($node->multipleType == 'or')
            {
                $this->dao->update(TABLE_APPROVALNODE)->set('status')->eq('done')->set('result')->eq('ignore')
                    ->where('approval')->eq($approvalID)
                    ->andWhere('node')->eq($node->node)
                    ->andWhere('status')->eq('doing')
                    ->exec();
            }
            else
            {
                $undone = $this->dao->select('*')->from(TABLE_APPROVALNODE)
                    ->where('approval')->eq($approvalID)
                    ->andWhere('node')->eq($node->node)
                    ->andWhere('type')->eq('review')
                    ->andWhere('status')->ne('done')
                    ->fetchAll();
                if(!empty($undone)) continue;
            }

            /* All reviewer is passed, run cc. */
            $ccs = $this->dao->select('account')->from(TABLE_APPROVALNODE)
                ->where('approval')->eq($approvalID)
                ->andWhere('node')->eq($node->node)
                ->andWhere('type')->eq('cc')
                ->fetchAll();

            $ccers = array();
            foreach($ccs as $cc) $ccers[] = $cc->account;

            $this->cc($approvalID, $node->node, $ccers);
            $this->next($approvalID, $node->node);
        }

        /* If the flow is finished, change status of approval. */
        $doing = $this->dao->select('*')->from(TABLE_APPROVALNODE)
            ->where('approval')->eq($approvalID)
            ->andWhere('status')->eq('doing')
            ->fetchAll();

        if(empty($doing))
        {
            $reject = $this->dao->select('*')->from(TABLE_APPROVALNODE)
                ->where('approval')->eq($approvalID)
                ->andWhere('result')->eq('fail')
                ->fetchAll();
            return $this->finish($approvalID, empty($reject) ? 'pass' : 'fail');
        }
    }

    /**
     * Reject approval.
     *
     * @param  string $objectType
     * @param  object $object
     * @param  string $extra
     * @access public
     * @return bool
     */
    public function reject($objectType = '', $object = null, $extra = '')
    {
        $approvalID = $this->dao->select('approval')->from(TABLE_APPROVALOBJECT)
            ->where('objectType')->eq($objectType)
            ->andWhere('objectID')->eq($object->id)
            ->orderBy('id_desc')
            ->fetch('approval');

        $data = new stdclass();
        $data->status       = 'done';
        $data->result       = 'fail';
        $data->date         = $object->createdDate;
        $data->opinion      = $object->opinion;
        $data->extra        = $extra;
        $data->reviewedBy   = $this->app->user->account;
        $data->reviewedDate = helper::now();

        $this->dao->update(TABLE_APPROVALNODE)->data($data)
            ->where('approval')->eq($approvalID)
            ->andWhere('status')->eq('doing')
            ->andWhere('account')->eq($this->app->user->account)
            ->exec();
        $this->dao->update(TABLE_APPROVALNODE)
            ->set('status')->eq('done')
            ->set('result')->eq('ignore')
            ->where('approval')->eq($approvalID)
            ->andWhere('status')->ne('done')
            ->exec();

        return $this->finish($approvalID, 'fail');
    }

    /**
     * Finish approval.
     *
     * @param int    $approvalID
     * @param string $result
     * @access public
     * @return bool
     */
    public function finish($approvalID, $result)
    {
        $this->dao->update(TABLE_APPROVAL)
            ->set('status')->eq('done')
            ->set('result')->eq($result)
            ->where('id')->eq($approvalID)
            ->exec();

        return true;
    }

    /**
     * Restart a approval flow.
     *
     * @param  string $objectType
     * @param  int    $objectID
     * @access public
     * @return void
     */
    public function restart($objectType, $objectID)
    {
        $approval = $this->dao->select('*')->from(TABLE_APPROVAL)
            ->where('objectType')->eq($objectType)
            ->andWhere('objectID')->eq($objectID)
            ->orderBy('id_desc')
            ->fetch();

        $approval->status = 'doing';
        $oldApprovalID    = $approval->id;
        unset($approval->id);

        $this->dao->insert(TABLE_APPROVAL)->data($approval)->exec();
        $approvalID = $this->dao->lastInsertID();

        $approvalObject = new stdclass();
        $approvalObject->approval   = $approvalID;
        $approvalObject->objectType = $objectType;
        $approvalObject->objectID   = $objectID;

        $this->dao->insert(TABLE_APPROVALOBJECT)->data($approvalObject)->exec();

        $nodes = $this->dao->select('*')->from(TABLE_APPROVALNODE)
            ->where('approval')->eq($oldApprovalID)
            ->fetchAll('id');

        foreach($nodes as $node)
        {
            unset($node->id);
            $node->status     = 'wait';
            $node->result     = '';
            $node->reviewedBy = '';
            $node->approval   = $approvalID;

            $this->dao->insert(TABLE_APPROVALNODE)->data($node)->exec();
        }

        $this->next($approvalID, '');
    }

    /**
     * Cancel all approval nodes of an object.
     *
     * @param  string $objectType
     * @param  int    $objectID
     * @access public
     * @return bool
     */
    public function cancel($objectType, $objectID)
    {
        $approvalID = $this->dao->select('approval')->from(TABLE_APPROVALOBJECT)
            ->where('objectType')->eq($objectType)
            ->andWhere('objectID')->eq($objectID)
            ->orderBy('id_desc')
            ->limit(1)
            ->fetch('approval');

        $this->dao->update(TABLE_APPROVALNODE)
            ->set('status')->eq('done')
            ->set('result')->eq('ignore')
            ->where('approval')->eq($approvalID)
            ->andWhere('status')->ne('done')
            ->exec();

        return !dao::isError();
    }

    /**
     * Build review desc.
     *
     * @param  object $node
     * @param  array  $extra
     * @access public
     * @return string
     */
    public function buildReviewDesc($node, $extra = array())
    {
        if(empty($node->id)) return '';

        $approval  = zget($extra, 'approval', '');
        $users     = zget($extra, 'users', array());
        $reviewers = zget($extra, 'reviewers', array());
        $allReviewers = zget($extra, 'allReviewers', array());
        $reviewDesc   = '';

        if($node->type == 'branch' and isset($node->branches))
        {
            foreach($node->branches as $childNode)
            {
                foreach($childNode->nodes as $branchChildNode) $reviewDesc .= $this->buildReviewDesc($branchChildNode, array('users' => $users, 'allReviewers' => $allReviewers, 'reviewers' => zget($allReviewers, $branchChildNode->id, array())));
            }
            return $reviewDesc;
        }

        if($node->type == 'start')
        {
            $reviewDesc  = "<li class='start' data-id='{$node->id}'><div><span class='timeline-text'>";
            $reviewDesc .= str_replace(array('$actor', '$date'), array(zget($users, $approval->createdBy), helper::isZeroDate($approval->createdDate) ? '' : substr($approval->createdDate, 0, 16)), $this->lang->approval->reviewDesc->start);
            $reviewDesc .= "</span></div></li>";
            return $reviewDesc;
        }

        $hasTitle    = !empty($node->title);
        $reviewClass = $hasTitle ? 'reviewer' : 'node';
        $reviewDesc  = '';

        $nodeStatus = 'wait';
        if(empty($reviewers) and isset($node->agentType) and $node->agentType == 'pass') $nodeStatus = 'pass';
        if(!empty($reviewers)) $nodeStatus = empty($reviewers['result']) ? $reviewers['status'] : $reviewers['result'];
        if($nodeStatus == 'success') $nodeStatus = 'pass';
        if($nodeStatus == 'ignore') return;

        if($hasTitle)
        {
            $reviewDesc .= "<li class='node title text-muted $nodeStatus' data-id='{$node->id}' data-status='{$nodeStatus}'><div><span class='timeline-text'>";
            $reviewDesc .= $node->title;
            if($node->multiple == 'or')$reviewDesc .= $this->lang->approval->notice->orSign;
            $reviewDesc .= "</span></div></li>";
        }

        if(empty($reviewers) and isset($node->agentType) and $node->agentType == 'pass')
        {
            $reviewDesc .= "<li class='$reviewClass $nodeStatus' data-id='{$node->id}' data-status='{$nodeStatus}'><div><span class='timeline-text'>";
            $reviewDesc .= $this->lang->approval->reviewDesc->pass4NoReviewer;
            $reviewDesc .= "</span></div></li>";
        }
        elseif(empty($reviewers['reviewers']) and !empty($reviewers['ccs']))
        {
            $ccList = '';
            foreach($reviewers['ccs'] as $cc) $ccList .= zget($users, $cc) . ' ';
            $reviewDesc .= "<li class='$reviewClass $nodeStatus' data-id='{$node->id}' data-status='{$nodeStatus}'><div><span class='timeline-text'>";
            $reviewDesc .= str_replace('$actor', trim($ccList), $this->lang->approval->reviewDesc->cc);
            $reviewDesc .= "</span></div></li>";
        }
        elseif(!empty($reviewers['reviewers']))
        {
            foreach($reviewers['reviewers'] as $reviewInfo)
            {
                $reviewStatus = !empty($reviewInfo['result']) ? $reviewInfo['result'] : $reviewInfo['status'];
                if(!isset($this->lang->approval->reviewDesc->$reviewStatus)) continue;

                if(!empty($reviewInfo['result']) and isset($node->multiple) and $node->multiple == 'or')
                {
                    /* Only show reviewed account in or node. */
                    if($reviewInfo['result'] != 'ignore')
                    {
                        $reviewDesc .= "<li class='$reviewClass $reviewStatus' data-id='{$node->id}' data-status='{$reviewStatus}'}><div><span class='timeline-text'>";
                        $reviewDesc .= str_replace(array('$actor', '$date'), array(zget($users, $reviewInfo['account']), helper::isZeroDate($reviewInfo['reviewedDate']) ? '' : substr($reviewInfo['reviewedDate'], 0, 16)), $this->lang->approval->reviewDesc->{$reviewStatus});
                        $reviewDesc .= "</span>";
                        if(!empty($reviewInfo['opinion'])) $reviewDesc .= "<div>{$reviewInfo['opinion']}</div>";
                        $reviewDesc .= "</div></li>";
                        break;
                    }
                }
                else
                {
                    if($reviewStatus == 'ignore') continue;

                    $account = $reviewInfo['account'];
                    $reviewDesc .= "<li class='$reviewClass $reviewStatus' data-id='{$node->id}' data-status='{$reviewStatus}'}><div><span class='timeline-text'>";
                    if(empty($account))
                    {
                        $autoReviewDesc = '';
                        if($node->reviewType == 'reject') $autoReviewDesc = $this->lang->approval->reviewDesc->autoReject;
                        if($node->reviewType == 'pass')   $autoReviewDesc = $this->lang->approval->reviewDesc->autoPass;
                        if(!empty($reviewInfo['result']) and $node->reviewType == 'reject') $autoReviewDesc = $this->lang->approval->reviewDesc->autoRejected;
                        if(!empty($reviewInfo['result']) and $node->reviewType == 'pass')   $autoReviewDesc = $this->lang->approval->reviewDesc->autoPassed;
                        $reviewDesc .= $autoReviewDesc;
                    }
                    else
                    {
                        $reviewDesc .= str_replace(array('$actor', '$date'), array(zget($users, $reviewInfo['account']), helper::isZeroDate($reviewInfo['reviewedDate']) ? '' : substr($reviewInfo['reviewedDate'], 0, 16)), $this->lang->approval->reviewDesc->{$reviewStatus});
                    }
                    $reviewDesc .= "</span>";
                    if(!empty($reviewInfo['opinion'])) $reviewDesc .= "<div class='opinion'>{$reviewInfo['opinion']}</div>";
                    $reviewDesc .= "</div></li>";
                }
            }
        }
        return $reviewDesc;
    }

    /**
     * Order branch nodes. order rule is done, doing, wait.
     *
     * @param  array    $nodes
     * @param  array    $reviewers
     * @access public
     * @return array
     */
    public function orderBranchNodes($nodes, $reviewers)
    {
        foreach($nodes as $i => $node)
        {
            if($node->type != 'branch') continue;
            if(count((array)$node->branches) <= 1) continue;

            $doneNodes  = array();
            $doingNodes = array();
            $waitNodes  = array();
            foreach($node->branches as $branckKey => $branchNodes)
            {
                $nodeStatus = 'wait';
                foreach($branchNodes->nodes as $branchNode)
                {
                    if($nodeStatus == 'doing') break;
                    if(!isset($reviewers[$branchNode->id]))
                    {
                        if(isset($branchNode->agentType) and $branchNode->agentType == 'pass') $nodeStatus = 'done';
                        continue;
                    }

                    if($reviewers[$branchNode->id]['status'] == 'doing') $nodeStatus = 'doing';
                    if($reviewers[$branchNode->id]['status'] == 'done')  $nodeStatus = 'done';
                    if($reviewers[$branchNode->id]['status'] == 'wait' and $nodeStatus == 'done') $nodeStatus = 'doing';
                }

                if($nodeStatus == 'done')  $doneNodes[]  = $branchNodes;
                if($nodeStatus == 'doing') $doingNodes[] = $branchNodes;
                if($nodeStatus == 'wait')  $waitNodes[]  = $branchNodes;
            }

            $sortedNodes = array();
            foreach($doneNodes as $branchNodes)  $sortedNodes[] = $branchNodes;
            foreach($doingNodes as $branchNodes) $sortedNodes[] = $branchNodes;
            foreach($waitNodes as $branchNodes)  $sortedNodes[] = $branchNodes;

            $node->branches = $sortedNodes;
        }

        return $nodes;
    }

    /*
     * Get approval node by approval id.
     *
     * @param  int    $approvalID
     * @param  string $type
     * @access public
     */
    public function getApprovalNodeByApprovalID($approvalID, $type = 'review')
    {
        return $this->dao->select('t1.id,t1.approval,t1.type,t1.result,t1.date,t1.opinion,t1.extra,t1.reviewedBy,t1.reviewedDate')->from(TABLE_APPROVALNODE)->alias('t1')
            ->where('type')->eq($type)
            ->andWhere('approval')->eq($approvalID)
            ->andWhere('result')->in('fail,pass')
            ->orderBy('reviewedDate asc')
            ->fetchAll();
    }

      /**
     * Build menu of actions created by workflow action.
     *
     * @param  string $module
     * @param  int    $data
     * @access public
     * @return string
     */
    public function buildApprovalMenu($module, $data)
    {
        if(!isset($this->config->bizVersion)) return '';

        $menu = '';
        $this->loadModel('flow');
        $moduleName = $module;
        $flow = $this->loadModel('workflow', 'flow')->getByModule($moduleName);
        if($flow->approval == 'enabled' && isset($data->approval) && !empty($data->approval))
        {
          
            if(strpos($module, '.') !== false) list($appName, $moduleName) = explode('.', $module);
        
            $extraClass = strpos(',testsuite,build,release,productplan,', ",{$moduleName},") !== false ? 'btn-link' : '';
    
            $menu .= baseHTML::a(helper::createLink('approval', 'progress', "approvalID={$data->approval}", '', true), "<i class='icon-spinner-indicator'></i>", "class='btn iframe' title='{$this->lang->flow->approvalProgress}'");
        }
       
        return $menu;
    }
}
