<?php
/**
 * The model file of approvalflow module of ZenTaoPMS.
 *
 * @copyright   Copyright 2009-2020 青岛易软天创网络科技有限公司(QingDao Nature Easy Soft Network Technology Co,LTD, www.cnezsoft.com)
 * @license     ZPL(http://zpl.pub/page/zplv12.html) or AGPL(https://www.gnu.org/licenses/agpl-3.0.en.html)
 * @author      Qiyu Xie <xieqiyu@easycorp.ltd>
 * @package     approvalflow
 * @version     $Id: model.php 5107 2020-09-09 09:46:12Z xieqiyu@easycorp.ltd $
 * @link        http://www.zentao.net
 */
class approvalflowModel extends model
{
    /**
     * My info
     *
     * @var array
     * @access private
     */
    private $submitter = array();

    /**
     * Get flow by id.
     *
     * @param  int $flowID
     * @param  int $version
     * @access public
     * @return void
     */
    public function getByID($flowID, $version = 0)
    {
        $flow = $this->dao->select('*')->from(TABLE_APPROVALFLOW)->where('id')->eq($flowID)->fetch();
        if(!$flow) return false;

        $spec = $this->dao->select('*')->from(TABLE_APPROVALFLOWSPEC)
            ->where('flow')->eq($flow->id)
            ->andWhere('version')->eq($version == 0 ? $flow->version : $version)
            ->fetch();

        $flow->version  = $spec->version;
        $flow->nodes    = $spec->nodes;
        $flow->workflow = '';

        if($flow->type == 'workflow')
        {
            $workflowiModule = $this->getWorkflowModule($flowID);
            $flow->workflow = $this->dao->select('name')->from(TABLE_WORKFLOW)->where('module')->eq($workflowiModule)->orderBy('id_desc')->fetch('name');
        }

        return $flow;
    }

    /**
     * Get flows.
     *
     * @param  string $type
     * @param  string $orderBy
     * @access public
     * @return array
     */
    public function getList($type = 'all', $orderBy = 'id_desc')
    {
        $flows = $this->dao->select('*')->from(TABLE_APPROVALFLOW)
            ->where('deleted')->eq(0)
            ->beginIF($type != 'all')->andWhere('type')->eq($type)->fi()
            ->orderBy($orderBy)
            ->fetchAll();

        $workflows = $this->dao->select('module,name')->from(TABLE_WORKFLOW)->fetchPairs('module');

        foreach($flows as $flow)
        {
            $flow->workflow = '';
            if($flow->type == 'workflow')
            {
                $workflowiModule = $this->getWorkflowModule($flow->id);
                $flow->workflow  = $workflows[$workflowiModule];
            }
        }

        return $flows;
    }

    /**
     * Get flow pairs.
     *
     * @param  string $type
     * @access public
     * @return void
     */
    public function getPairs($type = 'all')
    {
        $flows = $this->getList($type);

        if(empty($flows)) return array();

        $pairs = array();
        foreach($flows as $flow) $pairs[$flow->id] = $flow->name;

        return $pairs;
    }

    /**
     * Get approval flow id by object.
     *
     * @param  int    $rootID
     * @param  string $objectType
     * @access public
     * @return void
     */
    public function getFlowIDByObject($rootID = 0, $objectType = '')
    {
        if(!$objectType) return 0;

        $this->app->loadLang('baseline');
        $baselineObjects = $this->lang->baseline->objectList;

        $flowID = $this->dao->select('flow')->from(TABLE_APPROVALFLOWOBJECT)
            ->where('root')->eq($rootID)
            ->andWhere('objectType')->eq($objectType)
            ->orderBy('id_desc')
            ->fetch('flow');

        /* Baseline object list default simple flow. */
        if(in_array($objectType, array_keys($baselineObjects)) and !$flowID) $flowID = $this->dao->select('id')->from(TABLE_APPROVALFLOW)->where('code')->eq('simple')->fetch('id');

        return $flowID ? $flowID : 0;
    }

    /**
     * Gen nodes for approval flow.
     *
     * @param object $nodes
     * @param array  $my
     * @param array  $users
     * @access public
     * @return void
     */
    public function genNodes(&$nodes, $users = array())
    {
        if($users)
        {
            $reviewers = $users['reviewers'];
            $upLevel   = $users['upLevel'];
        }

        foreach($nodes as $node)
        {
            if($node->type == 'branch')
            {
                $isContinue = true; // Need execute other branch ?
                foreach($node->branches as $index => $branch)
                {
                    if($this->checkCondition($branch->conditions) and $isContinue)
                    {
                        $this->genNodes($branch->nodes, $users);
                        $node->branches[$index] = $branch;
                        if($node->branchType != 'parallel')
                        {
                            $isContinue = false;
                        }
                    }
                    else
                    {
                        unset($node->branches[$index]);
                    }
                }

                if($isContinue)
                {
                    $this->genNodes($node->default->nodes, $users);
                    $node->branches[] = $node->default;
                }
                unset($node->default);
            }
            else
            {
                $node->id       = in_array($node->type, array('start', 'end')) ? $node->type : $node->id;
                $node->title    = isset($node->title) ? $node->title : $this->lang->approvalflow->nodeTypeList[$node->type];
                $node->multiple = isset($node->multiple) ? $node->multiple : 'and';
                if($node->type == 'approval' and isset($node->reviewType) and $node->reviewType != 'manual')
                {
                    $reviewer = new stdclass();
                    $reviewer->users = array('');
                    $node->reviewers = array($reviewer);
                }
                else if(isset($node->reviewers) and !empty($node->reviewers))
                {
                    foreach($node->reviewers as $index => $reviewer)
                    {
                        if($reviewer->type == 'select')
                        {
                            if(!isset($reviewers[$node->id]) or !isset($reviewers[$node->id]['reviewers']))
                            {
                                dao::$errors[] = $this->lang->approvalflow->errorList['needReivewer'];
                                return false;
                            }
                            $node->reviewers[$index]->users = isset($reviewers) ? $reviewers[$node->id]['reviewers'] : array();
                        }
                        else if($reviewer->type == 'role')
                        {
                            $node->reviewers[$index]->users = array();
                            foreach($reviewer->roles as $role) $node->reviewers[$index]->users = array_merge($node->reviewers[$index]->users, $users['roles'][$role]);
                        }
                        else if($reviewer->type == 'projectRole')
                        {
                            $node->reviewers[$index]->users = array();
                            foreach($reviewer->projectRoles as $projectRole)
                            {
                                if(isset($users['projectRoles'][$projectRole])) $node->reviewers[$index]->users = array_merge($node->reviewers[$index]->users, $users['projectRoles'][$projectRole]);
                            }
                        }
                        else if($reviewer->type == 'upLevel')
                        {
                            $node->reviewers[$index]->users = $upLevel ? array($upLevel) : array();
                        }

                        $node->reviewers[$index]->users = array_unique($node->reviewers[$index]->users);
                    }

                    /* If reviewers is empty, use agent. */
                    if(!isset($node->reviewers[$index]->users) or empty($node->reviewers[$index]->users))
                    {
                        switch($node->agentType)
                        {
                            case 'appointee':
                                $node->reviewers[$index]->users = array($node->agentUser);
                                break;
                            case 'admin':
                                $admins = explode(',', trim($this->app->company->admins, ','));
                                $node->reviewers[$index]->users = array($admins[0]);
                                break;
                            default:
                                $node->reviewers[$index]->users = array();
                                break;
                        }
                    }
                }

                if(isset($node->ccs) and !empty($node->ccs))
                {
                    foreach($node->ccs as $index => $cc)
                    {
                        if($cc->type == 'select')
                        {
                            if(!isset($reviewers[$node->id]) or !isset($reviewers[$node->id]['ccs']))
                            {
                                dao::$errors[] = $this->lang->approvalflow->errorList['needCcer'];
                                return false;
                            }
                            $node->ccs[$index]->users = isset($reviewers) ? $reviewers[$node->id]['ccs'] : array();
                        }
                        else if($cc->type == 'role')
                        {
                            $node->ccs[$index]->users = array();
                            foreach($cc->roles as $role) $node->ccs[$index]->users = array_merge($node->ccs[$index]->users, $users['roles'][$role]);
                        }
                        else if($cc->type == 'projectRole')
                        {
                            $node->ccs[$index]->users = array();
                            foreach($cc->projectRoles as $projectRole)
                            {
                                if(isset($users['projectRoles'][$projectRole])) $node->ccs[$index]->users = array_merge($node->ccs[$index]->users, $users['projectRoles'][$projectRole]);
                            }
                        }
                        else if($cc->type == 'upLevel')
                        {
                            $node->ccs[$index]->users = $upLevel ? array($upLevel) : array();
                        }

                        $node->ccs[$index]->users = array_unique($node->ccs[$index]->users);
                    }
                }
            }
        }
    }

    /**
     * Get role list.
     *
     * @access public
     * @return array
     */
    public function getRoleList()
    {
        return $this->dao->select('*')->from(TABLE_APPROVALROLE)->where('deleted')->eq('0')->fetchAll('id');
    }

    /**
     * Get role pairs.
     *
     * @access public
     * @return array
     */
    public function getRolePairs()
    {
        return $this->dao->select('id,name')->from(TABLE_APPROVALROLE)->where('deleted')->eq('0')->fetchPairs();
    }

    /**
     * Create flow.
     *
     * @access public
     * @return void
     */
    public function create()
    {
        $data = new stdclass();
        $data->name        = $this->post->name;
        $data->type        = $this->post->type;
        $data->desc        = $this->post->desc;
        $data->createdBy   = $this->app->user->account;
        $data->createdDate = helper::now();

        $workflow = $this->post->workflow ? $this->post->workflow : '';

        if($data->type == 'workflow' and !$workflow)
        {
            dao::$errors['workflow'] = sprintf($this->lang->error->notempty, $this->lang->approvalflow->workflow);
            return false;
        }

        $this->dao->insert(TABLE_APPROVALFLOW)->data($data)
            ->autoCheck()
            ->batchCheck($this->config->approvalflow->create->requiredFields, 'notempty')
            ->exec();

        if(!dao::isError())
        {
            $flowID = $this->dao->lastInsertID();

            $spec = new stdclass();
            $spec->flow        = $flowID;
            $spec->version     = 1;
            $spec->nodes       = '[{"type":"start","ccs":[]},{"id":"3ewcj92p55e","type":"approval","reviewType":"manual","reviewers":[{"type":"select"}]},{"type":"end","ccs":[]}]';
            $spec->createdBy   = $this->app->user->account;
            $spec->createdDate = helper::now();

            $this->dao->insert(TABLE_APPROVALFLOWSPEC)->data($spec)->exec();

            if($data->type == 'workflow') $this->loadModel('workflow')->enableApproval($workflow, $flowID);

            return $flowID;
        }

        return false;
    }

    /**
     * Update flow.
     *
     * @param  int    $flowID
     * @access public
     * @return array
     */
    public function update($flowID)
    {
        $this->loadModel('workflow');

        $flow        = fixer::input('post')->get();
        $oldFlow     = $this->getByID($flowID);
        $oldWorkflow = $this->getWorkflowModule($flowID);
        $workflow    = $flow->type == 'workflow' ? $flow->workflow : '';
        unset($flow->workflow);

        if($oldWorkflow and $oldWorkflow != $workflow)
        {
            $result = $this->workflow->disableApproval($oldWorkflow);
            if($result['result'] == 'fail')
            {
                dao::$errors[] = $result['message'];
                return false;
            }
        }

        if($flow->type == 'workflow' and !$workflow)
        {
            dao::$errors['workflow'] = sprintf($this->lang->error->notempty, $this->lang->approvalflow->workflow);
            return false;
        }

        $this->dao->update(TABLE_APPROVALFLOW)->data($flow)
            ->where('id')->eq($flowID)
            ->batchCheck($this->config->approvalflow->edit->requiredFields, 'notempty')
            ->exec();

        if($flow->type == 'workflow' and $oldWorkflow != $workflow) $this->loadModel('workflow')->enableApproval($workflow, $flowID);

        return common::createChanges($oldFlow, $flow);
    }

    /**
     * Update nodes
     *
     * @param object $flow
     * @access public
     * @return void
     */
    public function updateNodes($flow)
    {
        $version = $flow->version + 1;

        $data = new stdclass();
        $data->flow        = $flow->id;
        $data->version     = $version;
        $data->nodes       = $this->post->nodes;
        $data->createdBy   = $this->app->user->account;
        $data->createdDate = helper::now();

        $this->dao->insert(TABLE_APPROVALFLOWSPEC)->data($data)->exec();

        $this->dao->update(TABLE_APPROVALFLOW)
            ->set('version')->eq($version)
            ->where('id')->eq($flow->id)
            ->exec();
    }

    /**
     * Create a approval role.
     *
     * @access public
     * @return void
     */
    public function createRole()
    {
        $data = fixer::input('post')
            ->join('users', ',')
            ->get();
        if($data->users) $data->users = ',' . trim($data->users, ',') . ',';

        $this->lang->approvalrole = new stdclass();
        $this->lang->approvalrole->name = $this->lang->approvalflow->name;
        $this->dao->insert(TABLE_APPROVALROLE)->data($data)->batchCheck('name', 'notempty')->exec();

        return $this->dao->lastInsertID();
    }

    /**
     * Edit a approval role.
     *
     * @access public
     * @return void
     */
    public function editRole($roleID)
    {
        $data = fixer::input('post')
            ->join('users', ',')
            ->get();
        if($data->users) $data->users = ',' . trim($data->users, ',') . ',';

        $this->lang->approvalrole = new stdclass();
        $this->lang->approvalrole->name = $this->lang->approvalflow->name;
        $this->dao->update(TABLE_APPROVALROLE)
            ->data($data)
            ->batchCheck('name', 'notempty')
            ->where('id')->eq($roleID)
            ->exec();

        return !dao::isError();
    }

    /**
     * Check condition
     *
     * @param  array $conditions
     * @access public
     * @return bool
     */
    public function checkCondition($conditions)
    {
        if(empty($conditions)) return true;

        if(empty($this->submitter))
        {
            /* Depts. */
            $path = '';
            if($this->app->user->dept) $path = $this->dao->select('path')->from(TABLE_DEPT)->where('id')->eq($this->app->user->dept)->fetch('path');
            $depts = explode(',', trim($path, ','));

            /* Roles. */
            $roles = $this->dao->select('id')->from(TABLE_APPROVALROLE)->where('users')->like('%,' . $this->app->user->account . ',%')->fetchAll('id');
            $roles = array_keys($roles);

            $this->submitter['account'] = $this->app->user->account;
            $this->submitter['depts']   = $depts;
            $this->submitter['roles']   = $roles;
        }

        foreach($conditions as $condition)
        {
            if($condition->type == 'user')
            {
                if($condition->selectType == 'account')
                {
                    if(in_array($this->app->user->account, $condition->users)) return true;
                }
                else if($condition->selectType == 'dept')
                {
                    foreach($this->submitter['depts'] as $dept)
                    {
                        if(in_array($dept, $condition->depts)) return true;
                    }
                }
                else if($condition->selectType == 'role')
                {
                    foreach($this->submitter['roles'] as $role)
                    {
                        if(in_array($role, $condition->roles)) return true;
                    }
                }
            }
        }

        return false;
    }

    /**
     * Search nodes to confirm.
     *
     * @param array  $nodes
     * @access public
     * @return array
     */
    public function searchNodesToConfirm($nodes, $module = '', $dataID = '')
    {
        $results = array();
        foreach($nodes as $node)
        {
            if($node->type == 'branch')
            {
                $exeDefault = true; // Need execute default branch ?
                foreach($node->branches as $branch)
                {
                    if(!$this->checkCondition($branch->conditions)) continue;

                    $results = array_merge($results, $this->searchNodesToConfirm($branch->nodes));
                    if($node->branchType != 'parallel')
                    {
                        $exeDefault = false;
                        break;
                    }
                }
                if($exeDefault) $results = array_merge($results, $this->searchNodesToConfirm($node->default->nodes));
            }
            else
            {
                $result = array('types' => array());
                if(in_array($node->type, array('start', 'end')))
                {
                    $result['id']    = $node->type;
                    $result['title'] = $this->lang->approvalflow->nodeTypeList[$node->type];
                }
                else
                {
                    $result['id']    = $node->id;
                    $result['title'] = isset($node->title) ? $node->title : $this->lang->approvalflow->nodeTypeList[$node->type];
                }

                if(isset($node->reviewers) and !empty($node->reviewers))
                {
                    $users  = array();
                    foreach($node->reviewers as $reviewer)
                    {
                        if(!isset($reviewer->type)) continue;
                        if($reviewer->type == 'select')
                        {
                            $result['types'][] = 'reviewer';
                            break;
                        }

                        if($reviewer->type == 'appointee')
                        {
                            foreach($reviewer->users as $account) $result['appointees']['reviewer'][] = $account;
                            break;
                        }

                        if($reviewer->type == 'projectRole')
                        { 
                            $table  = $this->dao->select('`table`')->from(TABLE_WORKFLOW)->where('module')->eq($module)->fetch('table') ?: '';
                            $object = $this->dao->select('*')->from($table)->where('id')->eq($dataID)->fetch();
                            if(!isset($object->project)) continue;
                            $project = $object->project;
                            foreach($reviewer->projectRoles as $projectRole)
                            {
                                $projectAccounts = $this->dao->select('*')->from(TABLE_TEAM)
                                    ->where('projectRole')->eq($projectRole)
                                    ->andWhere('root')->eq($project)
                                    ->fetchGroup('account');
                                $accounts = array_keys($projectAccounts);
                                $users    = array_unique(array_merge($users, $accounts));   
                            } 
                            $result['appointees']['reviewer'] = $users;
                            continue;
                        }

                        if($reviewer->type == 'role')
                        { 
                            $accounts = $this->dao->select('*')->from(TABLE_APPROVALROLE)->where('id')->in($reviewer->roles)->fetchAll(); 

                            foreach($accounts as $account)  $users = array_unique(array_merge($users, array_filter(explode(',' , $account->users ?: array()))));            
                            $result['appointees']['reviewer'] = $users;
                            continue;
                        } 
                    }
                }
        
                if(isset($node->ccs) and !empty($node->ccs))
                {
                    $users  = array();
                    foreach($node->ccs as $cc)
                    {
                        if(!isset($cc->type)) continue;
                        if($cc->type == 'select')
                        {
                            $result['types'][] = 'ccer';
                            break;
                        }

                        if($cc->type == 'appointee')
                        {
                            foreach($cc->users as $account) $result['appointees']['ccer'][] = $account;
                            break;
                        }

                        if($reviewer->type == 'projectRole')
                        {           
                            $table  = $this->dao->select('`table`')->from(TABLE_WORKFLOW)->where('module')->eq($module)->fetch('table') ?: '';
                            $object = $this->dao->select('*')->from($table)->where('id')->eq($dataID)->fetch();
                            if(!isset($object->project)) continue;
                            $project = $object->project;
                            foreach($cc->projectRoles as $ccRole)
                            {
                                $projectAccounts = $this->dao->select('*')->from(TABLE_TEAM)
                                    ->where('projectRole')->eq($ccRole)
                                    ->andWhere('root')->eq($project)
                                    ->fetchGroup('account');
                                $accounts = array_keys($projectAccounts);
                                $users    = array_unique(array_merge($users, $accounts));           
                            } 
                            $result['appointees']['ccer'] = $users;
                            continue;
                        }

                        if($reviewer->type == 'role')
                        { 
                            $accounts = $this->dao->select('*')->from(TABLE_APPROVALROLE)->where('id')->in($cc->roles)->fetchAll(); 

                            foreach($accounts as $account)  $users = array_unique(array_merge($users, array_filter(explode(',' , $account->users ?: array()))));            
                            $result['appointees']['ccer'] = $users;
                            continue;
                        } 
                    }
                }

                if(count($result['types']) >= 1 or isset($result['appointees'])) $results[] = $result;
            }
        }
        return $results;
    }

    /**
     * Adjust the action is clickable.
     *
     * @param  object  $flow
     * @param  string  $action
     *
     * @access public
     * @return bool
     */
    public static function isClickable($issue, $action)
    {
        $action = strtolower($action);

        if($action == 'delete') return !$issue->code;
        if($action == 'edit')   return !$issue->code;

        return true;
    }

    /**
     * Save layout.
     *
     * @param  string $module
     * @param  string $action
     * @access public
     * @return bool
     */
    public function saveLayout($module, $action, $flowID)
    {
        $fields = $this->loadModel('workflowfield')->getList($module);

        foreach($this->post->show as $field => $show)
        {
            $defaultValue = isset($this->post->defaultValue[$field]) ? $this->post->defaultValue[$field] : '';

            if($defaultValue)
            {
                $fieldInfo = $fields[$field];
                $fieldInfo->default = $defaultValue;
                $result = $this->loadModel('workflowfield')->checkDefaultValue($fieldInfo);

                if(is_array($result) && $result['result'] == 'fail' && is_array($result['message'])) dao::$errors["defaultValue$field"] = $result['message']['default'];
                if(is_array($result) && $result['result'] == 'fail' && is_string($result['message'])) dao::$errors["defaultValue$field"] = $result['message'];
            }
        }

        if(!empty(dao::$errors)) return false;

        $this->dao->delete()->from(TABLE_WORKFLOWNODELAYOUT)
            ->where('module')->eq($module)
            ->andWhere('action')->eq($action)
            ->andWhere('flow')->eq($flowID)
            ->beginIF(!empty($this->config->vision))->andWhere('vision')->eq($this->config->vision)->fi()
            ->exec();

        $order  = 1;
        $layout = new stdclass();
        $layout->module = $module;
        $layout->action = $action;
        foreach($this->post->show as $field => $show)
        {
            if(!$show) continue;

            /* Check width validate. */
            if(isset($this->post->width[$field]) && filter_var($this->post->width[$field], FILTER_VALIDATE_INT) === false && $this->post->width[$field] != 'auto')
            {
                dao::$errors['width' . $field] = sprintf($this->lang->error->int[0], $this->lang->workflowlayout->width);
                return false;
            }

            $defaultValue = isset($this->post->defaultValue[$field]) ? $this->post->defaultValue[$field] : '';
            if(is_array($defaultValue)) $defaultValue = implode(',', array_values(array_unique(array_filter($defaultValue))));

            $summary = isset($this->post->summary[$field]) ? $this->post->summary[$field] : '';
            if(is_array($summary)) $summary = implode(',', $summary);

            $layout->flow         = $flowID;
            $layout->field        = $field;
            $layout->order        = $order++;
            $layout->width        = (isset($this->post->width[$field]) && $this->post->width[$field] != 'auto' && $this->post->width[$field] != '') ? $this->post->width[$field] : 0;
            $layout->position     = isset($this->post->position[$field])   ? $this->post->position[$field]   : '';
            $layout->readonly     = isset($this->post->readonly[$field])   ? $this->post->readonly[$field]   : '0';
            $layout->mobileShow   = isset($this->post->mobileShow[$field]) ? $this->post->mobileShow[$field] : '0';
            $layout->summary      = $summary;
            $layout->defaultValue = $defaultValue;
            $layout->layoutRules  = isset($this->post->layoutRules[$field]) ? implode(',', $this->post->layoutRules[$field]) : '';
            if(!empty($this->config->vision)) $layout->vision = $this->config->vision;

            $this->dao->insert(TABLE_WORKFLOWNODELAYOUT)->data($layout)->autoCheck()->exec();
        }

        return !dao::isError();
    }
}
